@@ -112,49 +112,67 @@ void remove_socket_events(event_handler * handler, socket_event_source const* co
 	handler->event_loop_.filter_events(socket_event_filter);
 }
 
-void change_socket_event_handler(event_handler * old_handler, event_handler * new_handler, socket_event_source const* const source)
+socket_event_flag change_socket_event_handler(event_handler * old_handler, event_handler * new_handler, socket_event_source const* const source, socket_event_flag remove)
 {
 	if (!old_handler) {
-		return;
-	}
-
-	if (old_handler == new_handler) {
-		return;
+		return socket_event_flag{};
 	}
 
 	if (!new_handler) {
 		remove_socket_events(old_handler, source);
+		return socket_event_flag{};
 	}
-	else {
-		auto socket_event_filter = [&](event_loop::Events::value_type & ev) -> bool {
-			if (ev.first == old_handler) {
-				if (ev.second->derived_type() == socket_event::type()) {
-					if (std::get<0>(static_cast<socket_event const&>(*ev.second).v_) == source) {
-						ev.first = new_handler;
-					}
-				}
-				else if (ev.second->derived_type() == hostaddress_event::type()) {
-					if (std::get<0>(static_cast<hostaddress_event const&>(*ev.second).v_) == source) {
-						ev.first = new_handler;
+
+	// Don't check for old == new as we need to remove the events in remove and report the pending ones.
+
+#if DEBUG_SOCKETEVENTS
+	socket_event_flag seen = socket_event_flag{};
+#endif
+
+	socket_event_flag ret = socket_event_flag{};
+	auto socket_event_filter = [&](event_loop::Events::value_type & ev) -> bool {
+		if (ev.first == old_handler) {
+			if (ev.second->derived_type() == socket_event::type()) {
+				auto & sev = static_cast<socket_event const&>(*ev.second);
+				if (std::get<0>(sev.v_) == source) {
+					auto const flag = std::get<1>(sev.v_);
+#if DEBUG_SOCKETEVENTS
+					assert(!(flag & seen));
+					seen |= flag;
+#endif
+					if (flag & remove) {
+						return true;
 					}
+					ret |= flag;
+					ev.first = new_handler;
 				}
 			}
-			return false;
-		};
+			else if (ev.second->derived_type() == hostaddress_event::type()) {
+				if (std::get<0>(static_cast<hostaddress_event const&>(*ev.second).v_) == source) {
+					ev.first = new_handler;
+				}
+			}
+		}
+		return false;
+	};
 
-		old_handler->event_loop_.filter_events(socket_event_filter);
-	}
+	old_handler->event_loop_.filter_events(socket_event_filter);
+	return ret;
 }
 
-namespace {
-bool has_pending_event(event_handler * handler, socket_event_source const* const source, socket_event_flag event)
+#if DEBUG_SOCKETEVENTS
+bool FZ_PRIVATE_SYMBOL has_pending_event(event_handler * handler, socket_event_source const* const source, socket_event_flag event)
 {
+	if (!handler) {
+		return false;
+	}
+
 	bool ret = false;
 
 	auto socket_event_filter = [&](event_loop::Events::value_type const& ev) -> bool {
 		if (ev.first == handler && ev.second->derived_type() == socket_event::type()) {
 			auto const& socket_ev = static_cast<socket_event const&>(*ev.second).v_;
-			if (std::get<0>(socket_ev) == source && std::get<1>(socket_ev) == event) {
+			if (std::get<0>(socket_ev) == source && std::get<1>(socket_ev) & event) {
 				ret = true;
 			}
 		}
@@ -165,7 +183,9 @@ bool has_pending_event(event_handler * handler, socket_event_source const* const
 
 	return ret;
 }
+#endif
 
+namespace {
 #ifdef FZ_WINDOWS
 static int convert_msw_error_code(int error)
 {
@@ -666,7 +686,7 @@ protected:
 			}
 
 			// We're now interested in all the other nice events
-			waiting_ |= WAIT_READ | WAIT_WRITE;
+			waiting_ |= WAIT_READ;
 
 			return 1;
 		}
@@ -1045,6 +1065,7 @@ protected:
 		if (!socket_ || !socket_->evt_handler_) {
 			return;
 		}
+
 		if (triggered_ & WAIT_READ) {
 			socket_->evt_handler_->send_event<socket_event>(socket_->ev_source_, socket_event_flag::read, triggered_errors_[1]);
 			triggered_ &= ~WAIT_READ;
@@ -1206,23 +1227,6 @@ void socket_base::detach_thread(scoped_lock & l)
 	}
 }
 
-void socket_base::do_set_event_handler(event_handler* pEvtHandler)
-{
-	if (!socket_thread_) {
-		return;
-	}
-
-	scoped_lock l(socket_thread_->mutex_);
-
-	if (evt_handler_ == pEvtHandler) {
-		return;
-	}
-
-	change_socket_event_handler(evt_handler_, pEvtHandler, ev_source_);
-
-	evt_handler_ = pEvtHandler;
-}
-
 int socket_base::close()
 {
 	if (!socket_thread_) {
@@ -1244,7 +1248,7 @@ int socket_base::close()
 		static_cast<socket*>(this)->state_ = socket_state::closed;
 	}
 	else {
-		static_cast<listen_socket*>(this)->state_ = listen_socket_state::none;
+		static_cast<listen_socket*>(this)->state_ = listen_socket_state{};
 	}
 
 	socket_thread_->triggered_ = 0;
@@ -1427,7 +1431,7 @@ listen_socket::listen_socket(thread_pool & pool, event_handler* evt_handler)
 
 listen_socket::~listen_socket()
 {
-	if (state_ != listen_socket_state::none) {
+	if (state_ != listen_socket_state{}) {
 		close();
 	}
 
@@ -1435,9 +1439,26 @@ listen_socket::~listen_socket()
 	detach_thread(l);
 }
 
+void listen_socket::set_event_handler(event_handler* pEvtHandler)
+{
+	if (!socket_thread_) {
+		return;
+	}
+
+	scoped_lock l(socket_thread_->mutex_);
+
+	if (evt_handler_ == pEvtHandler) {
+		return;
+	}
+
+	change_socket_event_handler(evt_handler_, pEvtHandler, ev_source_, fz::socket_event_flag{});
+
+	evt_handler_ = pEvtHandler;
+}
+
 int listen_socket::listen(address_type family, int port)
 {
-	if (state_ != listen_socket_state::none) {
+	if (state_ != listen_socket_state{}) {
 		return EALREADY;
 	}
 
@@ -1514,7 +1535,7 @@ int listen_socket::listen(address_type family, int port)
 	socket_thread_->waiting_ = WAIT_ACCEPT;
 
 	if (socket_thread_->start()) {
-		state_ = listen_socket_state::none;
+		state_ = listen_socket_state{};
 		close_socket_fd(fd_);
 		return EMFILE;
 	}
@@ -1522,14 +1543,14 @@ int listen_socket::listen(address_type family, int port)
 	return 0;
 }
 
-std::unique_ptr<socket> listen_socket::accept(int &error)
+std::unique_ptr<socket> listen_socket::accept(int &error, fz::event_handler * handler)
 {
 	socket_descriptor desc = fast_accept(error);
 	if (!desc) {
 		return std::unique_ptr<socket>();
 	}
 
-	auto ret = socket::from_descriptor(std::move(desc), thread_pool_, error);
+	auto ret = socket::from_descriptor(std::move(desc), thread_pool_, error, handler);
 	if (!ret) {
 		error = ENOMEM;
 	}
@@ -1579,7 +1600,7 @@ socket_descriptor listen_socket::fast_accept(int &error)
 listen_socket_state listen_socket::get_state() const
 {
 	if (!socket_thread_) {
-		return listen_socket_state::none;
+		return listen_socket_state{};
 	}
 
 	scoped_lock l(socket_thread_->mutex_);
@@ -1603,7 +1624,7 @@ socket::~socket()
 	detach_thread(l);
 }
 
-std::unique_ptr<socket> socket::from_descriptor(socket_descriptor && desc, thread_pool & pool, int & error)
+std::unique_ptr<socket> socket::from_descriptor(socket_descriptor && desc, thread_pool & pool, int & error, fz::event_handler * handler)
 {
 	if (!desc) {
 		error = ENOTSOCK;
@@ -1629,7 +1650,8 @@ std::unique_ptr<socket> socket::from_descriptor(socket_descriptor && desc, threa
 		pSocket->state_ = socket_state::connected;
 		pSocket->fd_ = fd;
 		pSocket->host_ = to_native(pSocket->peer_ip());
-		pSocket->socket_thread_->waiting_ = WAIT_READ | WAIT_WRITE;
+		pSocket->evt_handler_ = handler;
+		pSocket->socket_thread_->waiting_ = WAIT_READ;
 		if (pSocket->socket_thread_->start()) {
 			error = ENOMEM;
 			pSocket.reset();
@@ -1641,7 +1663,7 @@ std::unique_ptr<socket> socket::from_descriptor(socket_descriptor && desc, threa
 
 int socket::connect(native_string const& host, unsigned int port, address_type family)
 {
-	if (state_ != socket_state::none) {
+	if (state_ != socket_state{}) {
 		return EISCONN;
 	}
 
@@ -1687,7 +1709,7 @@ int socket::connect(native_string const& host, unsigned int port, address_type f
 socket_state socket::get_state() const
 {
 	if (!socket_thread_) {
-		return socket_state::none;
+		return socket_state{};
 	}
 
 	scoped_lock l(socket_thread_->mutex_);
@@ -1701,6 +1723,14 @@ int socket::read(void* buffer, unsigned int size, int& error)
 		return -1;
 	}
 
+#if DEBUG_SOCKETEVENTS
+	{
+		scoped_lock l(socket_thread_->mutex_);
+		assert(!(socket_thread_->waiting_ & WAIT_READ));
+		assert(!has_pending_event(evt_handler_, this, socket_event_flag::read));
+	}
+#endif
+
 	int res = recv(fd_, (char*)buffer, size, 0);
 
 	if (res == -1) {
@@ -1735,6 +1765,14 @@ int socket::write(void const* buffer, unsigned int size, int& error)
 	int signal_set = sigaction(SIGPIPE, &action, &old_action);
 #endif
 
+#endif
+
+#if DEBUG_SOCKETEVENTS
+	if (socket_thread_) {
+		scoped_lock l(socket_thread_->mutex_);
+		assert(!(socket_thread_->waiting_ & WAIT_WRITE));
+		assert(!has_pending_event(evt_handler_, this, socket_event_flag::write | socket_event_flag::connection));
+	}
 #endif
 
 	int res = send(fd_, (const char*)buffer, size, flags);
@@ -1750,6 +1788,11 @@ int socket::write(void const* buffer, unsigned int size, int& error)
 		error = last_socket_error();
 		if (error == EAGAIN) {
 			scoped_lock l (socket_thread_->mutex_);
+
+#if DEBUG_SOCKETEVENTS
+			assert(!(socket_thread_->waiting_ & WAIT_WRITE));
+			assert(!has_pending_event(evt_handler_, this, socket_event_flag::write | socket_event_flag::connection));
+#endif
 			if (!(socket_thread_->waiting_ & WAIT_WRITE)) {
 				socket_thread_->waiting_ |= WAIT_WRITE;
 				socket_thread_->wakeup_thread(l);
@@ -1839,33 +1882,6 @@ native_string socket::peer_host() const
 	return host_;
 }
 
-void socket::retrigger(socket_event_flag event)
-{
-	if (!socket_thread_) {
-		return;
-	}
-
-	if (event != socket_event_flag::read && event != socket_event_flag::write) {
-		return;
-	}
-
-	scoped_lock l(socket_thread_->mutex_);
-
-	auto s = state_;
-	if (s != socket_state::connected && (s != socket_state::shut_down || event == socket_event_flag::write)) {
-		return;
-	}
-
-	if (!evt_handler_ || has_pending_event(evt_handler_, this, event)) {
-		return;
-	}
-
-	int const wait_flag = (event == socket_event_flag::read) ? WAIT_READ : WAIT_WRITE;
-	if (!(socket_thread_->waiting_ & wait_flag)) {
-		evt_handler_->send_event<socket_event>(this, event, 0);
-	}
-}
-
 int socket::shutdown()
 {
 #ifdef FZ_WINDOWS
@@ -1887,16 +1903,24 @@ int socket::shutdown()
 	return 0;
 }
 
-void socket::set_event_handler(event_handler* pEvtHandler)
+void socket::set_event_handler(event_handler* pEvtHandler, fz::socket_event_flag retrigger_block)
 {
-	do_set_event_handler(pEvtHandler);
+	if (!socket_thread_) {
+		return;
+	}
+
+	scoped_lock l(socket_thread_->mutex_);
+
+	fz::socket_event_flag const pending = change_socket_event_handler(evt_handler_, pEvtHandler, ev_source_, retrigger_block);
+	evt_handler_ = pEvtHandler;
 
 	if (pEvtHandler) {
-		scoped_lock l(socket_thread_->mutex_);
-		if (state_ == socket_state::connected && !(socket_thread_->waiting_ & WAIT_WRITE) && !has_pending_event(evt_handler_, ev_source_, socket_event_flag::write)) {
+		if (state_ == socket_state::connected && !(socket_thread_->waiting_ & WAIT_WRITE) && !(pending & (socket_event_flag::connection | socket_event_flag::write)) && !(retrigger_block & socket_event_flag::write)) {
+			socket_thread_->triggered_ &= ~WAIT_WRITE;
 			pEvtHandler->send_event<socket_event>(ev_source_, socket_event_flag::write, 0);
 		}
-		if ((state_ == socket_state::connected || state_ == socket_state::shut_down) && !(socket_thread_->waiting_ & WAIT_READ) && !has_pending_event(evt_handler_, ev_source_, socket_event_flag::read)) {
+		if ((state_ == socket_state::connected || state_ == socket_state::shut_down) && !(socket_thread_->waiting_ & WAIT_READ) && !(pending & socket_event_flag::read) && !(retrigger_block & socket_event_flag::read)) {
+			socket_thread_->triggered_ &= ~WAIT_READ;
 			pEvtHandler->send_event<socket_event>(ev_source_, socket_event_flag::read, 0);
 		}
 	}
@@ -1970,14 +1994,15 @@ socket_layer::~socket_layer()
 	remove_socket_events(event_handler_, this);
 }
 
-void socket_layer::set_event_handler(event_handler* handler)
+void socket_layer::set_event_handler(event_handler* handler, fz::socket_event_flag retrigger_block)
 {
 	auto old = event_handler_;
 	event_handler_ = handler;
-	change_socket_event_handler(old, handler, this);
+	socket_event_flag const pending = change_socket_event_handler(old, handler, this, retrigger_block);
 
 	if (event_passthrough_) {
-		next_layer_.set_event_handler(handler);
+		retrigger_block |= pending;
+		next_layer_.set_event_handler(handler, retrigger_block);
 	}
 }
 
@@ -1995,10 +2020,10 @@ void socket_layer::forward_hostaddress_event(socket_event_source* source, std::s
 	}
 }
 
-void socket_layer::set_event_passthrough()
+void socket_layer::set_event_passthrough(socket_event_flag retrigger_block)
 {
 	event_passthrough_ = true;
-	next_layer_.set_event_handler(event_handler_);
+	next_layer_.set_event_handler(event_handler_, retrigger_block);
 }
 
 int socket_layer::shutdown_read()
