@@ -15,6 +15,14 @@
 
 #include <string.h>
 
+#if DEBUG_SOCKETEVENTS
+#include <assert.h>
+
+namespace fz {
+bool FZ_PRIVATE_SYMBOL has_pending_event(event_handler * handler, socket_event_source const* const source, socket_event_flag event);
+}
+#endif
+
 static_assert(GNUTLS_VERSION_NUMBER != 0x030604, "Using TLS 1.3 with this version of GnuTLS does not work, update your version of GnuTLS");
 
 namespace fz {
@@ -138,6 +146,14 @@ struct cert_list_holder final
 	unsigned int certs_size{};
 };
 
+std::string to_string(gnutls_datum_t const& d)
+{
+	if (d.data && d.size) {
+		return std::string(d.data, d.data + d.size);
+	}
+	return {};
+}
+
 struct datum_holder final : gnutls_datum_t
 {
 	datum_holder() {
@@ -367,6 +383,11 @@ bool tls_layer_impl::init_session(bool client)
 	gnutls_transport_set_pull_function(session_, c_pull_function);
 	gnutls_transport_set_ptr(session_, (gnutls_transport_ptr_t)this);
 
+	if (!do_set_alpn()) {
+		deinit();
+		return false;
+	}
+
 	return true;
 }
 
@@ -459,10 +480,14 @@ void tls_layer_impl::log_alert(logmsg::type logLevel)
 	gnutls_alert_description_t last_alert = gnutls_alert_get(session_);
 	char const* alert = gnutls_alert_get_name(last_alert);
 	if (alert) {
-		logger_.log(logLevel, fztranslate("Received TLS alert from the server: %s (%d)"), alert, last_alert);
+		logger_.log(logLevel,
+					server_ ? fztranslate("Received TLS alert from the client: %s (%d)") : fztranslate("Received TLS alert from the server: %s (%d)"),
+					alert, last_alert);
 	}
 	else {
-		logger_.log(logLevel, fztranslate("Received unknown TLS alert %d from the server"), last_alert);
+		logger_.log(logLevel,
+					server_ ? fztranslate("Received unknown TLS alert %d from the client") : fztranslate("Received unknown TLS alert %d from the server"),
+					last_alert);
 	}
 }
 
@@ -578,6 +603,8 @@ void tls_layer_impl::on_socket_event(socket_event_source* s, socket_event_flag t
 		if (hostname_.empty()) {
 			set_hostname(tls_layer_.next_layer_.peer_host());
 		}
+		on_send();
+		break;
 	default:
 		break;
 	}
@@ -598,6 +625,10 @@ void tls_layer_impl::on_read()
 	}
 	else if (state_ == socket_state::connected || state_ == socket_state::shutting_down || state_ == socket_state::shut_down) {
 		if (tls_layer_.event_handler_) {
+#if DEBUG_SOCKETEVENTS
+			assert(!debug_can_read_);
+			debug_can_read_ = true;
+#endif
 			tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::read, 0);
 		}
 	}
@@ -665,9 +696,17 @@ int tls_layer_impl::continue_write()
 	}
 	while (!send_buffer_.empty());
 
-	if (state_ == socket_state::connected) {
-		if (tls_layer_.event_handler_) {
-			tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::write, 0);
+	if (write_blocked_by_send_buffer_) {
+		write_blocked_by_send_buffer_ = false;
+
+		if (state_ == socket_state::connected) {
+			if (tls_layer_.event_handler_) {
+#if DEBUG_SOCKETEVENTS
+				assert(!debug_can_write_);
+				debug_can_write_ = true;
+#endif
+				tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::write, 0);
+			}
 		}
 	}
 
@@ -738,6 +777,9 @@ bool tls_layer_impl::client_handshake(std::vector<uint8_t> const& session_to_res
 	if (!session_hostname.empty()) {
 		set_hostname(session_hostname);
 	}
+	else if (!hostname_.empty()) {
+		set_hostname(hostname_);
+	}
 
 	if (tls_layer_.next_layer_.get_state() == socket_state::none || tls_layer_.next_layer_.get_state() == socket_state::connecting) {
 		// Wait until the socket gets connected
@@ -754,7 +796,7 @@ bool tls_layer_impl::client_handshake(std::vector<uint8_t> const& session_to_res
 	return continue_handshake() == EAGAIN;
 }
 
-bool tls_layer_impl::server_handshake(std::vector<uint8_t> const& session_to_resume)
+bool tls_layer_impl::server_handshake(std::vector<uint8_t> const& session_to_resume, std::string_view const& preamble)
 {
 	logger_.log(logmsg::debug_verbose, L"tls_layer_impl::server_handshake()");
 
@@ -763,6 +805,8 @@ bool tls_layer_impl::server_handshake(std::vector<uint8_t> const& session_to_res
 		return false;
 	}
 
+	server_ = true;
+
 	ticket_key_ = session_to_resume;
 
 	if (!init() || !init_session(false)) {
@@ -784,6 +828,8 @@ bool tls_layer_impl::server_handshake(std::vector<uint8_t> const& session_to_res
 		return false;
 	}
 
+	preamble_.append(preamble);
+
 	return continue_handshake() == EAGAIN;
 }
 
@@ -794,6 +840,24 @@ int tls_layer_impl::continue_handshake()
 		return ENOTCONN;
 	}
 
+	while (!preamble_.empty()) {
+		if (!can_write_to_socket_) {
+			return EAGAIN;
+		}
+
+		int error{};
+		int written = tls_layer_.next_layer_.write(preamble_.get(), static_cast<int>(preamble_.size()), error);
+		if (written < 0) {
+			can_write_to_socket_ = false;
+			if (error != EAGAIN) {
+				socket_error_ = error;
+				failure(0, true);
+			}
+			return error;
+		}
+		preamble_.consume(static_cast<size_t>(written));
+	}
+
 	int res = gnutls_handshake(session_);
 	while (res == GNUTLS_E_AGAIN || res == GNUTLS_E_INTERRUPTED) {
 		if (!(gnutls_record_get_direction(session_) ? can_write_to_socket_ : can_read_from_socket_)) {
@@ -825,11 +889,16 @@ int tls_layer_impl::continue_handshake()
 			if (tls_layer_.event_handler_) {
 				tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::connection, 0);
 				if (can_read_from_socket_) {
+#if DEBUG_SOCKETEVENTS
+					assert(!debug_can_read_);
+					debug_can_read_ = true;
+#endif
 					tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::read, 0);
 				}
-				if (can_write_to_socket_) {
-					tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::write, 0);
-				}
+#if DEBUG_SOCKETEVENTS
+				assert(!debug_can_write_);
+				debug_can_write_ = true;
+#endif
 			}
 		}
 
@@ -861,12 +930,20 @@ int tls_layer_impl::read(void *buffer, unsigned int len, int& error)
 		return -1;
 	}
 
+#if DEBUG_SOCKETEVENTS
+	assert(debug_can_read_);
+	assert(!has_pending_event(tls_layer_.event_handler_, &tls_layer_, socket_event_flag::read));
+#endif
+
 	int res = do_call_gnutls_record_recv(buffer, len);
 	if (res >= 0) {
 		error = 0;
 		return res;
 	}
 	else if (res == GNUTLS_E_INTERRUPTED || res == GNUTLS_E_AGAIN) {
+#if DEBUG_SOCKETEVENTS
+		debug_can_read_ = false;
+#endif
 		error = EAGAIN;
 	}
 	else {
@@ -892,7 +969,16 @@ int tls_layer_impl::write(void const* buffer, unsigned int len, int& error)
 		return -1;
 	}
 
+#if DEBUG_SOCKETEVENTS
+	assert(debug_can_write_);
+	assert(!has_pending_event(tls_layer_.event_handler_, &tls_layer_, socket_event_flag::write));
+#endif
+
 	if (!send_buffer_.empty()) {
+		write_blocked_by_send_buffer_ = true;
+#if DEBUG_SOCKETEVENTS
+		debug_can_write_ = false;
+#endif
 		error = EAGAIN;
 		return -1;
 	}
@@ -944,7 +1030,7 @@ void tls_layer_impl::failure(int code, bool send_close, std::wstring const& func
 				)
 			{
 				if (state_ != socket_state::shut_down || !shutdown_silence_read_errors_) {
-					logger_.log(logmsg::status, fztranslate("Server did not properly shut down TLS connection"));
+					logger_.log(logmsg::status, server_ ? fztranslate("Client did not properly shut down TLS connection") : fztranslate("Server did not properly shut down TLS connection"));
 				}
 			}
 		}
@@ -1048,11 +1134,16 @@ void tls_layer_impl::set_verification_result(bool trusted)
 		if (tls_layer_.event_handler_) {
 			tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::connection, 0);
 			if (can_read_from_socket_) {
+#if DEBUG_SOCKETEVENTS
+				assert(!debug_can_read_);
+				debug_can_read_ = true;
+#endif
 				tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::read, 0);
 			}
-			if (can_write_to_socket_) {
-				tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::write, 0);
-			}
+#if DEBUG_SOCKETEVENTS
+			assert(!debug_can_write_);
+			debug_can_write_ = true;
+#endif
 		}
 
 		return;
@@ -1798,7 +1889,7 @@ std::string tls_layer_impl::get_gnutls_version()
 void tls_layer_impl::set_hostname(native_string const& host)
 {
 	hostname_ = host;
-	if (!hostname_.empty() && get_address_type(hostname_) == address_type::unknown) {
+	if (session_ && !hostname_.empty() && get_address_type(hostname_) == address_type::unknown) {
 		auto const utf8 = to_utf8(hostname_);
 		if (!utf8.empty()) {
 			int res = gnutls_server_name_set(session_, GNUTLS_NAME_DNS, utf8.c_str(), utf8.size());
@@ -1809,6 +1900,33 @@ void tls_layer_impl::set_hostname(native_string const& host)
 	}
 }
 
+native_string tls_layer_impl::get_hostname() const
+{
+	if (!session_) {
+		return {};
+	}
+
+	size_t len{};
+	unsigned int type{};
+	unsigned int i{};
+	int ret;
+	do {
+		ret = gnutls_server_name_get(session_, nullptr, &len, &type, i++);
+	}
+	while (ret == GNUTLS_E_SHORT_MEMORY_BUFFER && type != GNUTLS_NAME_DNS);
+
+	if (ret == GNUTLS_E_SHORT_MEMORY_BUFFER) {
+		std::string name;
+		name.resize(len - 1);
+		ret = gnutls_server_name_get(session_, name.data(), &len, &type, --i);
+		if (!ret) {
+			return fz::to_native(name);
+		}
+	}
+
+	return {};
+}
+
 int tls_layer_impl::connect(native_string const& host, unsigned int port, address_type family)
 {
 	if (hostname_.empty()) {
@@ -2005,4 +2123,61 @@ int tls_layer_impl::shutdown_read()
 	return error;
 }
 
+void tls_layer_impl::set_event_handler(event_handler* pEvtHandler, fz::socket_event_flag retrigger_block)
+{
+	write_blocked_by_send_buffer_ = false;
+
+	fz::socket_event_flag const pending = change_socket_event_handler(tls_layer_.event_handler_, pEvtHandler, &tls_layer_, retrigger_block);
+	tls_layer_.event_handler_ = pEvtHandler;
+
+	if (pEvtHandler) {
+		if ((state_ == socket_state::connected || state_ == socket_state::shutting_down) && !(pending & (socket_event_flag::write | socket_event_flag::connection)) && !(retrigger_block & socket_event_flag::write)) {
+			pEvtHandler->send_event<socket_event>(&tls_layer_, socket_event_flag::write, 0);
+		}
+		if ((state_ == socket_state::connected || state_ == socket_state::shutting_down || state_ == socket_state::shut_down) && !(pending & socket_event_flag::read) && !(retrigger_block & socket_event_flag::read)) {
+			pEvtHandler->send_event<socket_event>(&tls_layer_, socket_event_flag::read, 0);
+		}
+
+#if DEBUG_SOCKETEVENTS
+		if (state_ == socket_state::connected || state_ == socket_state::shutting_down || state_ == socket_state::shut_down) {
+			debug_can_read_ = true;
+		}
+		if (state_ == socket_state::connected || state_ == socket_state::shutting_down) {
+			debug_can_write_ = true;
+		}
+#endif
+	}
+
+}
+
+bool tls_layer_impl::do_set_alpn()
+{
+	if (alpn_.empty()) {
+		return true;
+	}
+
+	gnutls_datum_t * data = new gnutls_datum_t[alpn_.size()];
+	for (size_t i = 0; i < alpn_.size(); ++i) {
+		data[i].data = reinterpret_cast<unsigned char *>(const_cast<char*>(alpn_[i].c_str()));
+		data[i].size = alpn_[i].size();
+	}
+	int res = gnutls_alpn_set_protocols(session_, data, alpn_.size(), GNUTLS_ALPN_MANDATORY);
+
+	if (res) {
+		log_error(res, L"gnutls_alpn_set_protocols");
+	}
+	return res == 0;
+}
+
+std::string tls_layer_impl::get_alpn() const
+{
+	if (session_) {
+		gnutls_datum_t protocol;
+		if (!gnutls_alpn_get_selected_protocol(session_, &protocol)) {
+			return to_string(protocol);
+		}
+	}
+	return {};
+}
+
 }
