@@ -28,7 +28,7 @@ public:
 
 	bool client_handshake(std::vector<uint8_t> const& session_to_resume, native_string const& session_hostname, std::vector<uint8_t> const& required_certificate, event_handler * verification_handler);
 
-	bool server_handshake(std::vector<uint8_t> const& session_to_resume);
+	bool server_handshake(std::vector<uint8_t> const& session_to_resume, std::string_view const& preamble);
 
 	int connect(native_string const& host, unsigned int port, address_type family);
 
@@ -69,6 +69,11 @@ public:
 
 	int shutdown_read();
 
+	void set_event_handler(event_handler* pEvtHandler, fz::socket_event_flag retrigger_block);
+
+	std::string get_alpn() const;
+	native_string get_hostname() const;
+
 private:
 	bool init();
 	void deinit();
@@ -112,6 +117,8 @@ private:
 		return ticket_key_.empty();
 	}
 
+	bool do_set_alpn();
+
 	tls_layer& tls_layer_;
 
 	logger_interface & logger_;
@@ -122,6 +129,8 @@ private:
 
 	gnutls_certificate_credentials_t cert_credentials_{};
 
+	std::vector<std::string> alpn_;
+
 	socket_state state_{};
 
 	bool handshake_successful_{};
@@ -139,6 +148,9 @@ private:
 	// gnutls_record_get_max_size()
 	buffer send_buffer_;
 
+	// Sent out just before the handshake itself
+	buffer preamble_;
+
 	std::vector<uint8_t> required_certificate_;
 
 	friend class tls_layer;
@@ -154,6 +166,14 @@ private:
 	bool socket_eof_{};
 
 	bool initialized_{};
+	bool server_{};
+
+	bool write_blocked_by_send_buffer_{};
+
+#if DEBUG_SOCKETEVENTS
+	bool debug_can_read_{};
+	bool debug_can_write_{};
+#endif
 };
 }
 
